@@ -25,9 +25,9 @@
 
 package org.glite.voms;
 
+import java.io.ByteArrayInputStream;
 import java.io.IOException;
 import java.io.InputStream;
-import java.io.ByteArrayInputStream;
 import java.net.InetAddress;
 import java.net.UnknownHostException;
 import java.security.PublicKey;
@@ -39,26 +39,27 @@ import java.security.cert.CertificateNotYetValidException;
 import java.security.cert.X509CRL;
 import java.security.cert.X509CRLEntry;
 import java.security.cert.X509Certificate;
+import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Date;
 import java.util.Enumeration;
+import java.util.HashSet;
 import java.util.Hashtable;
 import java.util.Iterator;
+import java.util.List;
+import java.util.Map;
 import java.util.NoSuchElementException;
 import java.util.Set;
 import java.util.Stack;
 import java.util.TreeSet;
-import java.util.HashSet;
 import java.util.Vector;
 
 import javax.security.auth.x500.X500Principal;
 
 import org.apache.log4j.Logger;
-import org.bouncycastle.asn1.DERObjectIdentifier;
-import org.bouncycastle.asn1.ASN1Sequence;
-import org.bouncycastle.asn1.ASN1Object;
-import org.bouncycastle.asn1.DERObject;
 import org.bouncycastle.asn1.ASN1InputStream;
+import org.bouncycastle.asn1.DERObject;
+import org.bouncycastle.asn1.DERObjectIdentifier;
 import org.bouncycastle.asn1.DEROctetString;
 import org.bouncycastle.asn1.x509.X509Extension;
 import org.bouncycastle.asn1.x509.X509Extensions;
@@ -534,42 +535,24 @@ public class PKIVerifier {
 
         X509Certificate currentCert = certs[0];
 
-        logger.debug( "path length = " + certs.length );
-
         for ( int i = 1; i < certs.length; i++ ) {
             if ( logger.isDebugEnabled() )
                 logger.debug( "Checking: " + certs[i].getSubjectDN().getName() );
 
             if ( PKIUtils.checkIssued( certs[i], certs[i - 1] ) ) {
 
-                logger.debug( "Is issuer" );
-
-                if (PKIUtils.isProxy(currentCert)) {
-                    logger.debug("Increase proxycount");
+                if (PKIUtils.isProxy(currentCert))                     
                     proxyCount ++;
-                }
+                
 
                 certStack.push( certs[i] );
                 currentCert = certs[i];
-
-                if ( logger.isDebugEnabled() )
-                    logger.debug( "ELEMENT: "
-                            + currentCert.getSubjectDN().getName() );
             }
-            logger.debug( "Is not issuer" );
+            
         }
 
-        logger.debug( "Before anchor searching." );
         X509Certificate candidate = null;
 
-        if ( logger.isDebugEnabled() ) {
-            Iterator j = certStack.iterator();
-            while ( j.hasNext() )
-                logger.debug( "Content: "
-                        + ( (X509Certificate) j.next() ).getSubjectDN()
-                                .getName() );
-        }
-
         // replace self-signed certificate passed with one from store.
         if ( PKIUtils.selfIssued( currentCert ) ) {
             String hash = PKIUtils.getHash( currentCert );
@@ -581,46 +564,33 @@ public class PKIVerifier {
                 candidate = (X509Certificate) candidates.elementAt( index );
                 certStack.push( candidate );
 
-                if ( logger.isDebugEnabled() )
-                    logger.debug( "ELEMENT: "
-                            + candidate.getSubjectDN().getName() );
             } else {
                 logger.error("Cannot find issuer candidate for: " +currentCert.getSubjectDN().getName());
                 return false;
             }
         } else {
             candidate = null;
-
-            logger.debug( "Looking for anchor" );
             // now, complete the certification path.
             do {
                 String hash = PKIUtils.getHash( currentCert
                         .getIssuerX500Principal() );
 
-                logger.debug( "hash = " + hash );
+                logger.debug( "Issuer principal hash = " + hash );
                 Vector candidates = (Vector) certificates.get( hash );
                 if ( candidates != null ) {
 
-                    logger.debug( "CANDIDATES: " + candidates );
+                    logger.debug( "Candidates trust anchors from store: " + candidates );
                     Iterator i = candidates.iterator();
                     while ( i.hasNext() ) {
                         candidate = (X509Certificate) i.next();
 
                         if ( logger.isDebugEnabled() )
-                            logger.debug( "Candidate = "
+                            logger.debug( "Candidate trust anchor subject = "
                                     + candidate.getSubjectDN().getName() );
 
                         if ( PKIUtils.checkIssued( candidate, currentCert ) ) {
                             certStack.push( candidate );
                             currentCert = candidate;
-
-//                             if (PKIUtils.isProxy(candidate))
-//                                 proxyCount ++;
-
-                            if ( logger.isDebugEnabled() )
-                                logger.debug( "ELEMENT: "
-                                        + candidate.getSubjectDN().getName() );
-
                             break;
                         } else
                             candidate = null;
@@ -636,22 +606,22 @@ public class PKIVerifier {
         }
 
         int currentLength = 0;
-        boolean success = true;
 
         PublicKey candidatePublicKey = null;
         X509Certificate issuerCert = null;
 
         if ( logger.isDebugEnabled() ) {
-            logger.debug( "Constructed chain:" );
+            logger.debug( "Constructed certificate chain:" );
 
             Iterator j = certStack.iterator();
+            int chainIndex = 0;
             while ( j.hasNext() )
-                logger.debug( "Content: "
+                logger.debug( "["+chainIndex+"]: "
                         + ( (X509Certificate) j.next() ).getSubjectDN()
                                 .getName() );
         }
 
-        // now verifies it
+        
         int stackSize = certStack.size();
         int stackPos = stackSize+1;
         Stack constructedStack = new Stack();
@@ -663,9 +633,9 @@ public class PKIVerifier {
 
             if (!PKIUtils.isProxy(currentCert))
                 certCount ++;
-
+            
             if ( logger.isDebugEnabled() )
-                logger.debug( "VERIFYING : "
+                logger.debug( "Checking : "
                         + currentCert.getSubjectDN().getName() );
 
             if (!checkProxyCertInfo(currentCert, stackPos, stackSize)) {
@@ -675,20 +645,16 @@ public class PKIVerifier {
             if ( PKIUtils.selfIssued( currentCert ) ) {
                 if ( currentLength != 0 ) {
                     logger
-                            .error( "Certificate verification: Self signed certificate not trust anchor" );
-                    logger.error( "subject: "
-                            + currentCert.getSubjectDN().getName() );
-                    success = false;
-                    break;
+                            .error( "Certificate verification: Self signed certificate not trust anchor: " + 
+                    currentCert.getSubjectDN().getName()  );
+                    return false;
                 } else {
-                    // this is the trust anchor
+                    
                     candidatePublicKey = currentCert.getPublicKey();
                     issuerCert = currentCert;
                 }
             }
-
-            logger.debug( "Checking chain" );
-
+            
             if ( !currentCert.getIssuerX500Principal().equals(
                     issuerCert.getSubjectX500Principal() ) ) {
                 logger
@@ -696,8 +662,6 @@ public class PKIVerifier {
                 return false;
             }
 
-            logger.debug( "Checking validity" );
-
             try {
                 currentCert.checkValidity();
             } catch ( CertificateExpiredException e ) {
@@ -720,8 +684,6 @@ public class PKIVerifier {
                 return false;
             }
 
-            logger.debug( "Checking key" );
-
             try {
                 currentCert.verify( candidatePublicKey );
             } catch ( Exception e ) {
@@ -732,9 +694,7 @@ public class PKIVerifier {
                         + currentCert.getSubjectDN().getName() );
                 return false;
             }
-
-            logger.debug( "Checking revoked" );
-
+            
             if ( isRevoked( currentCert, issuerCert ) ) {
                 logger
                         .error( "Certificate verification: certificate in chain has been revoked." );
@@ -744,8 +704,6 @@ public class PKIVerifier {
             }
 
             boolean isCA = PKIUtils.isCA( issuerCert );
-
-            logger.debug( "Checking CA " + isCA );
             if ( isCA ) {
                 if ( !allowsPath( currentCert, issuerCert, constructedStack ) ) {
                     logger.error( "Certificate verification: subject '"
@@ -758,22 +716,16 @@ public class PKIVerifier {
                 // check path length
                 int maxPath = currentCert.getBasicConstraints();
 
-                logger.debug( "stack.size = " + certStack.size()
-                        + " maxPath = " + maxPath );
-
                 if ( maxPath != -1 ) {
                     if ( maxPath < certStack.size() - proxyCount - 1  ) {
                         logger
                                 .error( "Certificate verification: Maximum certification path length exceeded." );
-                        success = false;
-                        break;
+                        return false;
                     }
                 }
             } else {
-                // not a ca. Maybe a proxy?
-                logger.debug( "Checking for Proxy" );
-
-                if ( !PKIUtils.isProxy( currentCert ) ) {
+                
+            	if ( !PKIUtils.isProxy( currentCert ) ) {
                     logger
                             .error( "Certificate verification: Non-proxy, non-CA certificate issued a certificate." );
                     return false;
@@ -794,10 +746,7 @@ public class PKIVerifier {
             candidatePublicKey = currentCert.getPublicKey();
             currentLength++;
             constructedStack.push(currentCert);
-        }
-
-        if ( !success )
-            return false;
+        }        
 
         return true;
     }
@@ -961,66 +910,137 @@ public class PKIVerifier {
         return null;
     }
 
+    private boolean checkCRLIssuer(X509CRL crl, X509Certificate issuer){
+    	
+    	return crl.getIssuerX500Principal().equals(issuer.getSubjectX500Principal());
+    }
+    
+    private boolean checkCRLCriticalExtensions(X509CRL crl){
+    	
+    	// Check critical extensions
+		Set criticalExts = crl.getCriticalExtensionOIDs();
+		
+		Set permittedCriticals = new HashSet();
+		permittedCriticals.add("2.5.29.28");
+		
+		if (criticalExts == null || criticalExts.isEmpty())
+			return true;
+		
+		if (!criticalExts.isEmpty() && !criticalExts.containsAll(permittedCriticals)){
+			logger.error("CRL critical extensions check failed for CRL "+crl.getIssuerX500Principal()+". Critical extensions "+permittedCriticals+" not found!");
+			return false;
+		}
+		
+    	return true;
+    }
+    
+    
+    private boolean checkCRLValidity(X509CRL crl){
+    	
+    	Date now = new Date();
+    	return crl.getNextUpdate().after(now) && crl.getThisUpdate().before(now);
+    }
+    
+    /**
+     * Looks for CRLs that match a given CA certificate.
+     * 
+     * @param issuer
+     * @return <code>null</code> if no CRLs were found in the trust-anchors directory, or a list of valid
+     * CRLs. When the list is empty the caller can assume that the CRL was present in the trust-anchors but
+     * was ill-formed.
+     * 	
+     */
+    private List<X509CRL> lookupCRL(X509Certificate issuer) {
+    	Map<String, List<X509CRL>> crlMap = caStore.getCRLs();
+    	
+    	List<X509CRL> crlList =  crlMap.get(PKIUtils.getHash(issuer));
+    	
+    	List<X509CRL> correctCrls = new ArrayList<X509CRL>();
+    	
+    	if (crlList == null || crlList.isEmpty())
+    		return null;
+    	
+    	for (X509CRL candidateCRL: crlList){
+    		
+    		// Verify signature
+    		try{
+    			candidateCRL.verify(issuer.getPublicKey());
+    			
+    		}catch (Exception e){
+    			logger.info("Signature verification check failed for CRL "+candidateCRL.getIssuerX500Principal()+"...");
+    			continue;
+    		}
+    		
+    		
+    		boolean criticalExtensionsAreValid = checkCRLCriticalExtensions(candidateCRL);
+    		if (!criticalExtensionsAreValid){
+    			logger.info("Critical extensions check failed for CRL "+ candidateCRL.getIssuerX500Principal()+"...");
+    			continue;
+    		}
+    		
+    		if (!checkCRLIssuer(candidateCRL, issuer)){
+    			logger.info(String.format("Issuer check failed for CRL %s against issuer %s.", candidateCRL.getIssuerX500Principal(), issuer.getSubjectX500Principal()));
+    			continue;
+    		}
+    		correctCrls.add(candidateCRL);
+
+    	}
+    	
+    	return correctCrls;
+    }
+    
+    
     private boolean isRevoked( X509Certificate cert, X509Certificate issuer ) {
 
-        Hashtable crls = caStore.getCRLs();
-        Vector crllist = (Vector) ( crls.get( PKIUtils.getHash( issuer ) ) );
-
-        boolean issued = false;
-
-        if ( crllist != null ) {
-            Iterator i = crllist.iterator();
-
-            while ( i.hasNext() ) {
-
-                X509CRL candidateCRL = (X509CRL) i.next();
-
-                if ( candidateCRL != null ) {
-                    try {
-                        candidateCRL.verify( issuer.getPublicKey() );
-                    } catch ( Exception e ) {
-                        continue;
-                    }
-
-                    Set criticalExts = candidateCRL.getCriticalExtensionOIDs();
-
-                    Set permittedCriticals = new HashSet();
-                    permittedCriticals.add("2.5.29.28");
-
-                    if ( criticalExts == null || criticalExts.isEmpty() ||
-                         (!criticalExts.isEmpty() && permittedCriticals.containsAll(criticalExts))) {
-                        if ( candidateCRL.getIssuerX500Principal().equals(
-                                          issuer.getSubjectX500Principal() ) ) {
-                            if ( candidateCRL.getNextUpdate().compareTo(
-                                        new Date() ) >= 0
-                                 && candidateCRL.getThisUpdate()
-                                 .compareTo( new Date() ) <= 0 ) {
-
-                                X509CRLEntry entry = candidateCRL
-                                    .getRevokedCertificate( cert
-                                        .getSerialNumber() );
-                                if ( entry == null ) {
-                                    return false;
-                                }
-                            }
-                            else {
-                                logger.error( "CRL for CA '"+issuer.getSubjectDN().toString()+"' has expired!" );
-                                return true;
-                            }
-                        }
-                    }
-                    else {
-                        logger.error("Critical extension found in crl!");
-                        for (Iterator ii = criticalExts.iterator(); ii.hasNext();) {
-                            logger.debug("Critical CRL Extension: " +
-                                         (String)ii.next());
-                        }
-
-                    }
-                    issued = true;
-                }
-            }
-        }
-        return issued;
+    	
+    	List<X509CRL> crls = lookupCRL(issuer);
+    	
+    	if (crls == null){
+    		logger.warn("No CRL for CA '"+issuer.getSubjectDN()+"' was found in local trust-anchor dir. Considering certificate '"+cert.getSubjectDN()+"' valid.");
+    		return false;
+    	}
+    	
+    	if (crls.isEmpty()){
+    		logger.warn("CRLs for CA '"+issuer.getSubjectDN()+"' was found but was ill-formed. Considering the certificate '"+cert.getSubjectDN()+"' revoked.");
+    		return true;
+    	}
+    	
+    	X509CRL validCrl = null;
+    	
+    	for (X509CRL crl: crls){
+    	
+    		logger.debug("Checking CRL: "+crl);
+    	
+    		boolean crlIsValid  = checkCRLValidity(crl);
+    	
+    		logger.debug("CRL is valid? "+crlIsValid);
+    	
+    		// Break at first valid CRL found...
+    		if (crlIsValid){
+    			validCrl = crl;
+    			break;
+    		}
+    		
+    		String msg = String.format("CRL for CA '%s' has expired on %s.",
+    				issuer.getSubjectDN(), crl.getNextUpdate());
+    				
+    		logger.error(msg);
+    	}
+    			
+    	if (validCrl == null){
+    		logger.warn(" No temporally valid CRL for CA '"+issuer.getSubjectDN()+"' was found. Considering the certificate '"+cert.getSubjectDN()+"' revoked.");
+    		return true;
+    	}
+    		
+        X509CRLEntry entry = validCrl.getRevokedCertificate( cert.getSerialNumber() );
+        
+        logger.debug("CRLEntry for certificate serial number "+cert.getSerialNumber()+": "+entry);
+        
+        if (entry == null)
+        	return false;
+        
+        logger.info(String.format("Certificate %s (%d) was revoked on date %s.", cert.getSubjectDN(), entry.getSerialNumber(), entry.getRevocationDate()));
+        return true;
+        
     }
 }
